#-------------------------------------------------------------------------------
#
#  Copyright (c) 2007, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   06/06/2007
#
#-------------------------------------------------------------------------------

""" Class to aid in automatically computing the 'slice' points for a specified
    ImageResource and then drawing it that it can be 'stretched' to fit a larger
    region than the original image.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

import wx

from colorsys \
    import rgb_to_hls

from numpy \
    import reshape, fromstring, uint8

from traits.api \
    import HasPrivateTraits, Instance, Int, List, Color, Enum, Bool

from pyface.image_resource \
    import ImageResource

from constants \
    import WindowColor

from constants import is_mac
import traitsui.wx.constants

#-------------------------------------------------------------------------------
#  Recursively paint the parent's background if they have an associated image
#  slice.
#-------------------------------------------------------------------------------

def paint_parent ( dc, window ):
    """ Recursively paint the parent's background if they have an associated
        image slice.
    """
    parent = window.GetParent()
    slice  = getattr( parent, '_image_slice', None )
    if slice is not None:
        x, y   = window.GetPositionTuple()
        dx, dy = parent.GetSizeTuple()
        slice.fill( dc, -x, -y, dx, dy )
    else:
        # Otherwise, just paint the normal window background color:
        dx, dy = window.GetClientSizeTuple()
        if is_mac and hasattr(window, '_border') and window._border:
            dc.SetBackgroundMode(wx.TRANSPARENT)
            dc.SetBrush( wx.Brush( wx.Colour(0, 0, 0, 0)))
        else:
            dc.SetBrush( wx.Brush( parent.GetBackgroundColour() ) )
        dc.SetPen( wx.TRANSPARENT_PEN )
        dc.DrawRectangle( 0, 0, dx, dy )

    return slice

#-------------------------------------------------------------------------------
#  'ImageSlice' class:
#-------------------------------------------------------------------------------

class ImageSlice ( HasPrivateTraits ):

    #-- Trait Definitions ------------------------------------------------------

    # The ImageResource to be sliced and drawn:
    image = Instance( ImageResource )

    # The minimum number of adjacent, identical rows/columns needed to identify
    # a repeatable section:
    threshold = Int( 10 )

    # The maximum number of 'stretchable' rows and columns:
    stretch_rows    = Enum( 1, 2 )
    stretch_columns = Enum( 1, 2 )

    # Width/height of the image borders:
    top    = Int
    bottom = Int
    left   = Int
    right  = Int

    # Width/height of the extended image borders:
    xtop    = Int
    xbottom = Int
    xleft   = Int
    xright  = Int

    # The color to use for content text:
    content_color = Instance( wx.Colour )

    # The color to use for label text:
    label_color = Instance( wx.Colour )

    # The background color of the image:
    bg_color = Color

    # Should debugging slice lines be drawn?
    debug = Bool( False )

    #-- Private Traits ---------------------------------------------------------

    # The current image's opaque bitmap:
    opaque_bitmap = Instance( wx.Bitmap )

    # The current image's transparent bitmap:
    transparent_bitmap = Instance( wx.Bitmap )

    # Size of the current image:
    dx = Int
    dy = Int

    # Size of the current image's slices:
    dxs = List
    dys = List

    # Fixed minimum size of current image:
    fdx = Int
    fdy = Int

    #-- Public Methods ---------------------------------------------------------

    def fill ( self, dc, x, y, dx, dy, transparent = False ):
        """ 'Stretch fill' the specified region of a device context with the
            sliced image.
        """
        # Create the source image dc:
        idc = wx.MemoryDC()
        if transparent:
            idc.SelectObject( self.transparent_bitmap )
        else:
            idc.SelectObject( self.opaque_bitmap )

        # Set up the drawing parameters:
        sdx, sdy = self.dx, self.dx
        dxs, dys = self.dxs, self.dys
        tdx, tdy = dx - self.fdx, dy - self.fdy

        # Calculate vertical slice sizes to use for source and destination:
        n = len( dxs )
        if n == 1:
            pdxs = [ ( 0, 0 ), ( 1, max( 1, tdx/2 ) ), ( sdx - 2, sdx - 2 ),
                     ( 1, max( 1, tdx - (tdx/2) ) ), ( 0, 0 ) ]
        elif n == 3:
            pdxs = [ ( dxs[0], dxs[0] ), ( dxs[1], max( 0, tdx ) ), ( 0, 0 ),
                     ( 0, 0 ), ( dxs[2], dxs[2] ) ]
        else:
            pdxs = [ ( dxs[0], dxs[0] ), ( dxs[1], max( 0, tdx/2 ) ),
                     ( dxs[2], dxs[2] ), ( dxs[3], max( 0, tdx - (tdx/2) ) ),
                     ( dxs[4], dxs[4] ) ]

        # Calculate horizontal slice sizes to use for source and destination:
        n = len( dys )
        if n == 1:
            pdys = [ ( 0, 0 ), ( 1, max( 1, tdy/2 ) ), ( sdy - 2, sdy - 2 ),
                     ( 1, max( 1, tdy - (tdy/2) ) ), ( 0, 0 ) ]
        elif n == 3:
            pdys = [ ( dys[0], dys[0] ), ( dys[1], max( 0, tdy ) ), ( 0, 0 ),
                     ( 0, 0 ), ( dys[2], dys[2] ) ]
        else:
            pdys = [ ( dys[0], dys[0] ), ( dys[1], max( 0, tdy/2 ) ),
                     ( dys[2], dys[2] ), ( dys[3], max( 0, tdy - (tdy/2) ) ),
                     ( dys[4], dys[4] ) ]

        # Iterate over each cell, performing a stretch fill from the source
        # image to the destination window:
        last_x, last_y = x + dx, y + dy
        y0, iy0 = y, 0
        for idy, wdy in pdys:
            if y0 >= last_y:
                break

            if wdy != 0:
                x0, ix0 = x, 0
                for idx, wdx in pdxs:
                    if x0 >= last_x:
                        break

                    if wdx != 0:
                        self._fill( idc, ix0, iy0, idx, idy,
                                    dc,  x0,  y0,  wdx, wdy )
                        x0 += wdx
                    ix0 += idx
                y0 += wdy
            iy0 += idy

        if self.debug:
            dc.SetPen( wx.Pen( wx.RED ) )
            dc.DrawLine( x, y + self.top, last_x, y + self.top )
            dc.DrawLine( x, last_y - self.bottom - 1,
                         last_x, last_y - self.bottom - 1 )
            dc.DrawLine( x + self.left, y, x + self.left, last_y )
            dc.DrawLine( last_x - self.right - 1, y,
                         last_x - self.right - 1, last_y )

    #-- Event Handlers ---------------------------------------------------------

    def _image_changed ( self, image ):
        """ Handles the 'image' trait being changed.
        """
        # Save the original bitmap as the transparent version:
        self.transparent_bitmap = bitmap = \
            image.create_image().ConvertToBitmap()

        # Save the bitmap size information:
        self.dx = dx = bitmap.GetWidth()
        self.dy = dy = bitmap.GetHeight()

        # Create the opaque version of the bitmap:
        self.opaque_bitmap = wx.EmptyBitmap( dx, dy )
        mdc2 = wx.MemoryDC()
        mdc2.SelectObject( self.opaque_bitmap )
        mdc2.SetBrush( wx.Brush( WindowColor ) )
        mdc2.SetPen( wx.TRANSPARENT_PEN )
        mdc2.DrawRectangle( 0, 0, dx, dy )
        mdc = wx.MemoryDC()
        mdc.SelectObject( bitmap )
        mdc2.Blit( 0, 0, dx, dy, mdc, 0, 0, useMask = True )
        mdc.SelectObject(  wx.NullBitmap )
        mdc2.SelectObject( wx.NullBitmap )

        # Finally, analyze the image to find out its characteristics:
        self._analyze_bitmap()

    #-- Private Methods --------------------------------------------------------

    def _analyze_bitmap ( self ):
        """ Analyzes the bitmap.
        """
        # Get the image data:
        threshold = self.threshold
        bitmap    = self.opaque_bitmap
        dx, dy    = self.dx, self.dy
        image     = bitmap.ConvertToImage()

        # Convert the bitmap data to a numpy array for analysis:
        data = reshape( fromstring( image.GetData(), uint8 ), ( dy, dx, 3 ) )

        # Find the horizontal slices:
        matches  = []
        y, last  = 0, dy - 1
        max_diff = 0.10 * dx
        while y < last:
            y_data = data[y]
            for y2 in xrange( y + 1, dy ):
                if abs( y_data - data[y2] ).sum() > max_diff:
                    break

            n = y2 - y
            if n >= threshold:
                matches.append( ( y, n ) )

            y = y2

        n = len( matches )
        if n == 0:
            if dy > 50:
                matches = [ ( 0, dy ) ]
            else:
                matches = [ ( dy / 2, 1 ) ]
        elif n > self.stretch_rows:
            matches.sort( lambda l, r: cmp( r[1], l[1] ) )
            matches = matches[ : self.stretch_rows ]

        # Calculate and save the horizontal slice sizes:
        self.fdy, self.dys = self._calculate_dxy( dy, matches )

        # Find the vertical slices:
        matches  = []
        x, last  = 0, dx - 1
        max_diff = 0.10 * dy
        while x < last:
            x_data = data[:,x]
            for x2 in xrange( x + 1, dx ):
                if abs( x_data - data[:,x2] ).sum() > max_diff:
                    break

            n = x2 - x
            if n >= threshold:
                matches.append( ( x, n ) )

            x = x2

        n = len( matches )
        if n == 0:
            if dx > 50:
                matches = [ ( 0, dx ) ]
            else:
                matches = [ ( dx / 2, 1 ) ]
        elif n > self.stretch_columns:
            matches.sort( lambda l, r: cmp( r[1], l[1] ) )
            matches = matches[ : self.stretch_columns ]

        # Calculate and save the vertical slice sizes:
        self.fdx, self.dxs = self._calculate_dxy( dx, matches )

        # Save the border size information:
        self.top    = min( dy / 2, self.dys[0] )
        self.bottom = min( dy / 2, self.dys[-1] )
        self.left   = min( dx / 2, self.dxs[0] )
        self.right  = min( dx / 2, self.dxs[-1] )

        # Find the optimal size for the borders (i.e. xleft, xright, ... ):
        self._find_best_borders( data )

        # Save the background color:
        x, y          = (dx / 2), (dy / 2)
        r, g, b       = data[ y, x ]
        self.bg_color = (0x10000 * r) + (0x100 * g) + b

        # Find the best contrasting text color (black or white):
        self.content_color = self._find_best_color( data, x, y )

        # Find the best contrasting label color:
        if self.xtop >= self.xbottom:
            self.label_color = self._find_best_color( data, x, self.xtop / 2 )
        else:
            self.label_color = self._find_best_color(
                                        data, x, dy - (self.xbottom / 2) - 1 )

    def _fill ( self, idc, ix, iy, idx, idy, dc, x, y, dx, dy ):
        """ Performs a stretch fill of a region of an image into a region of a
            window device context.
        """
        last_x, last_y = x + dx, y + dy
        while y < last_y:
            ddy = min( idy, last_y - y )
            x0  = x
            while x0 < last_x:
                ddx = min( idx, last_x - x0 )
                dc.Blit( x0, y, ddx, ddy, idc, ix, iy, useMask = True )
                x0 += ddx
            y += ddy

    def _calculate_dxy ( self, d, matches ):
        """ Calculate the size of all image slices for a specified set of
            matches.
        """
        if len( matches ) == 1:
            d1, d2 = matches[0]

            return ( d - d2, [ d1, d2, d - d1 - d2 ] )

        d1, d2 = matches[0]
        d3, d4 = matches[1]

        return ( d - d2 - d4, [ d1, d2, d3 - d1 - d2, d4, d - d3 - d4 ] )

    def _find_best_borders ( self, data ):
        """ Find the best set of image slice border sizes (e.g. for images with
            rounded corners, there should exist a better set of borders than
            the ones computed by the image slice algorithm.
        """
        # Make sure the image size is worth bothering about:
        dx, dy = self.dx, self.dy
        if (dx < 5) or (dy < 5):
            return

        # Calculate the starting point:
        left = right  = dx / 2
        top  = bottom = dy / 2

        # Calculate the end points:
        last_y = dy - 1
        last_x = dx - 1

        # Mark which edges as 'scanning':
        t = b = l = r = True

        # Keep looping while at last one edge is still 'scanning':
        while l or r or t or b:

            # Calculate the current core area size:
            height = bottom - top + 1
            width  = right - left + 1

            # Try to extend all edges that are still 'scanning':
            nl = (l and (left > 0) and
                  self._is_equal( data, left - 1, top, left, top, 1, height ))

            nr = (r and (right < last_x) and
                  self._is_equal( data, right + 1, top, right, top, 1, height ))

            nt = (t and (top > 0) and
                 self._is_equal( data, left, top - 1, left, top, width, 1 ))

            nb = (b and (bottom < last_y) and
                  self._is_equal( data, left, bottom + 1, left, bottom,
                                  width, 1 ))

            # Now check the corners of the edges:
            tl = ((not nl) or (not nt) or
                  self._is_equal( data, left - 1, top - 1, left, top, 1, 1 ))

            tr = ((not nr) or (not nt) or
                  self._is_equal( data, right + 1, top - 1, right, top, 1, 1 ))

            bl = ((not nl) or (not nb) or
                  self._is_equal( data, left - 1, bottom + 1, left, bottom,
                                  1, 1 ))

            br = ((not nr) or (not nb) or
                  self._is_equal( data, right + 1, bottom + 1, right, bottom,
                                  1, 1 ))

            # Calculate the new edge 'scanning' values:
            l = nl and tl and bl
            r = nr and tr and br
            t = nt and tl and tr
            b = nb and bl and br

            # Adjust the coordinate of an edge if it is still 'scanning':
            left   -= l
            right  += r
            top    -= t
            bottom += b

        # Now compute the best set of image border sizes using the current set
        # and the ones we just calculated:
        self.xleft   = min( self.left,   left )
        self.xright  = min( self.right,  dx - right - 1 )
        self.xtop    = min( self.top,    top )
        self.xbottom = min( self.bottom, dy - bottom - 1 )

    def _find_best_color ( self, data, x, y ):
        """ Find the best contrasting text color for a specified pixel
            coordinate.
        """
        r, g, b = data[ y, x ]
        h, l, s = rgb_to_hls( r / 255.0, g / 255.0, b / 255.0 )
        text_color = wx.BLACK
        if l < 0.50:
            text_color = wx.WHITE

        return text_color

    def _is_equal ( self, data, x0, y0, x1, y1, dx, dy ):
        """ Determines if two identically sized regions of an image array are
            'the same' (i.e. within some slight color variance of each other).
        """
        return (abs( data[ y0: y0 + dy, x0: x0 + dx ] -
                     data[ y1: y1 + dy, x1: x1 + dx ] ).sum() < 0.10 * dx * dy)


#-------------------------------------------------------------------------------
#  Returns a (possibly cached) ImageSlice:
#-------------------------------------------------------------------------------

image_slice_cache = {}

def image_slice_for ( image ):
    """ Returns a (possibly cached) ImageSlice.
    """
    global image_slice_cache

    result = image_slice_cache.get( image )
    if result is None:
        image_slice_cache[ image ] = result = ImageSlice( image = image )

    return result



#-----------------------------------------------------------------------------
# Name:        bitviewscroller.py
# Purpose:     scrolling container for bit patterns
#
# Author:      Rob McMullen
#
# Created:     2014
# RCS-ID:      $Id: $
# Copyright:   (c) 2014 Rob McMullen
# License:     wxWidgets
#-----------------------------------------------------------------------------
"""BitviewScroller -- a container for viewing bit patterns

"""

import os
import time

import numpy as np
import wx
import wx.lib.newevent

from pyface.action.api import Action
from pyface.tasks.action.api import EditorAction

from atrcopy import SegmentData, DefaultSegment

from omnivore.tasks.hex_edit.actions import *

import logging
log = logging.getLogger(__name__)

class BitviewEvent(wx.PyCommandEvent):
    """Event sent when a LayerControl is changed."""

    def __init__(self, eventType, id, byte, bit):
        wx.PyCommandEvent.__init__(self, eventType, id)
        self.byte = byte
        self.bit = bit


class BitviewScroller(wx.ScrolledWindow):
    dbg_call_seq = 0
    short_name = "_bitview base class"
    
    def __init__(self, parent, task, **kwargs):
        wx.ScrolledWindow.__init__(self, parent, -1, **kwargs)

        # Settings
        self.task = task
        self.editor = None
        self.segment = None
        self.max_zoom = 16
        self.min_zoom = 1
        self.bytes_per_row = 1
        self.pixels_per_byte = 8
        self.bitplanes = 1

        # internal storage
        self.start_addr = 0
        self.start_byte = None
        self.end_byte = None
        self.img = None
        self.scaled_bmp = None
        self.grid_width = 0
        self.grid_height = 0
        self.zoom = 5
        self.start_row = 0
        self.start_col = 0
        self.fully_visible_rows = 1
        self.fully_visible_cols = 1
        self.visible_rows = 1
        self.visible_cols = 1

        self.rect_select = False
        
        self.select_extend_mode = False
        self.multi_select_mode = False
        self.Bind(wx.EVT_PAINT, self.on_paint)
        self.Bind(wx.EVT_SIZE, self.on_resize)
        self.Bind(wx.EVT_MOUSEWHEEL, self.on_mouse_wheel)
        self.Bind(wx.EVT_LEFT_UP, self.on_left_up)
        self.Bind(wx.EVT_LEFT_DOWN, self.on_left_down)
        self.Bind(wx.EVT_LEFT_DCLICK, self.on_left_dclick)
        self.Bind(wx.EVT_MOTION, self.on_motion)
        self.Bind(wx.EVT_RIGHT_DOWN, self.on_popup)
        self.Bind(wx.EVT_SET_FOCUS, self.on_focus)
        self.Bind(wx.EVT_KILL_FOCUS, self.on_focus_lost)
        self.Bind(wx.EVT_ENTER_WINDOW, self.on_mouse_enter)
        self.Bind(wx.EVT_LEAVE_WINDOW, self.on_mouse_leave)
        self.Bind(wx.EVT_CHAR, self.on_char)
        self.Bind(wx.EVT_CHAR_HOOK, self.on_char_hook)
    
    def __repr__(self):
        return "<%s at 0x%x>" % (self.__class__.__name__, id(self))
    
    def is_ready_to_render(self):
        return self.editor is not None
    
    def set_task(self, task):
        self.task = task
    
    def recalc_view(self):
        editor = self.task.active_editor
        if editor is not None:
            self.editor = editor
            self.segment = self.get_segment(editor)
            self.rect_select = editor.rect_select
            self.start_addr = editor.segment.start_addr
            self.update_bytes_per_row()
            self.set_colors()
            self.set_font()
            self.update_zoom()
            self.set_scale()

    def get_segment(self, editor):
        """Get segment from editor; provided for subclasses that use
        a computed segment to represent the segment data (like Jumpman)
        """
        return editor.segment
    
    def refresh_view(self):
        editor = self.task.active_editor
        if editor is not None:
            if self.editor != editor:
                self.recalc_view()
            else:
                self.Refresh()
        
    def sync_settings(self):
        e = self.editor
        if e is not None:
            self.sync_to_editor(e)
    
    def sync_to_editor(self, e):
        pass

    def set_colors(self):
        pass
    
    def set_font(self):
        pass

    def zoom_in(self, zoom=1):
        self.set_zoom(self.zoom + zoom)
        self.set_scale()
        
    def zoom_out(self, zoom=1):
        self.set_zoom(self.zoom - zoom)
        self.set_scale()
    
    def set_zoom(self, zoom):
        if zoom > self.max_zoom:
            zoom = self.max_zoom
        elif zoom < self.min_zoom:
            zoom = self.min_zoom
        self.zoom = zoom
        self.sync_settings()
    
    def get_zoom_factors(self):
        return self.zoom, self.zoom

    def get_highlight_indexes(self):
        e = self.editor
        anchor_start, anchor_end = e.anchor_start_index, e.anchor_end_index
        r1 = c1 = r2 = c2 = -1
        if self.rect_select:
            anchor_start, anchor_end, (r1, c1), (r2, c2) = self.segment.get_rect_indexes(anchor_start, anchor_end)
        elif anchor_start > anchor_end:
            anchor_start, anchor_end = anchor_end, anchor_start
        elif anchor_start == anchor_end:
            anchor_start = e.cursor_index
            anchor_end = anchor_start + 1
        return anchor_start, anchor_end, (r1, c1), (r2, c2)
    
    def get_image(self, segment=None):
        raise NotImplementedError

    def copy_to_clipboard(self):
        """Copies current image to clipboard.

        Copies the current image, including scaling, zooming, etc. to
        the clipboard.
        """
        bmpdo = wx.BitmapDataObject(self.scaled_bmp)
        if wx.TheClipboard.Open():
            wx.TheClipboard.SetData(bmpdo)
            wx.TheClipboard.Close()

    def prepare_image(self):
        """Creates new image at specified zoom factor.

        Creates a new image that is to be used as the background of
        the scrolled bitmap.  Currently, actually creates the entire
        image, which could lead to memory problems if the image is
        really huge and the zoom factor is large.
        """
        if self.is_ready_to_render():
            self.calc_image_size()
            
            w, h = self.GetClientSizeTuple()
            dc = wx.MemoryDC()
            self.scaled_bmp = wx.EmptyBitmap(w, h)
            
            dc.SelectObject(self.scaled_bmp)
            dc.SetBackground(wx.Brush(self.editor.machine.empty_color))
            dc.Clear()
            
            array = self.get_image()
            width = array.shape[1]
            height = array.shape[0]
            if width > 0 and height > 0:
                zw, zh = self.get_zoom_factors()
                nw = width * zw
                nh = height * zh
                newdims = np.asarray((nh, nw))
                base=np.indices(newdims)
                d = []
                d.append(base[0]/zh)
                d.append(base[1]/zw)
                cd = np.array(d)
                newarray = array[list(cd)]
                self.draw_overlay(newarray, width, height, zw, zh)
                image = wx.EmptyImage(nw, nh)
                image.SetData(newarray.tostring())
                bmp = wx.BitmapFromImage(image)
                dc.DrawBitmap(bmp, 0, 0)
    
    def draw_overlay(self, array, w, h, zw, zh):
        """ Hook to draw overlay image on top of finished scaled image
        """
        pass
    
    def calc_image_size(self):
        x, y = self.GetViewStart()
        w, h = self.GetClientSizeTuple()
        self.start_row = y
        self.start_col = x
        
        # For proper buffered paiting, the visible_rows must include the
        # (possibly) partially obscured last row.  fully_visible_rows
        # indicates the number of rows without that last partially obscured
        # row (if it exists).
        self.fully_visible_rows = h / self.zoom
        self.fully_visible_cols = w / self.zoom
        self.visible_rows = ((h + self.zoom - 1) / self.zoom)
        self.visible_cols = ((w + self.zoom - 1) / self.zoom)
        log.debug("x, y, w, h, start, num: %s" % str([x, y, w, h, self.start_row, self.visible_rows]))

    def update_bytes_per_row(self):
        self.pixels_per_byte = self.editor.machine.bitmap_renderer.pixels_per_byte
        self.bitplanes = self.editor.machine.bitmap_renderer.bitplanes

    def update_zoom(self):
        pass

    def set_scale(self):
        """Creates new image at specified zoom factor.

        Creates a new image that is to be used as the background of
        the scrolled bitmap.  Currently, actually creates the entire
        image, which could lead to memory problems if the image is
        really huge and the zoom factor is large.
        """
        if self.segment is not None:
            self.calc_scale_from_bytes()
        else:
            self.grid_width = 10
            self.grid_height = 10
        log.debug("set_scale: %s" % str([self.grid_width, self.grid_height]))
        self.calc_scroll_params()
        self.Refresh()
    
    def calc_scale_from_bytes(self):
        self.total_rows = (len(self.segment) + self.bytes_per_row - 1) / self.bytes_per_row
        self.grid_width = int(self.pixels_per_byte * self.bytes_per_row)
        self.grid_height = int(self.total_rows)
    
    def calc_scroll_params(self):
        self.SetVirtualSize((self.grid_width * self.zoom, self.grid_height * self.zoom))
        rate = int(self.zoom)
        if rate < 1:
            rate = 1
        self.SetScrollRate(rate, rate)

    def event_coords_to_byte(self, evt):
        """Convert event coordinates to world coordinates.

        Convert the event coordinates to world coordinates by locating
        the offset of the scrolled window's viewport and adjusting the
        event coordinates.
        """
        raise NotImplementedError
    
    def byte_to_row_col(self, addr):
        r = addr // self.bytes_per_row
        c = addr - (r * self.bytes_per_row)
        return r, c
    
    def select_index(self, rel_pos):
        r, c = self.byte_to_row_col(rel_pos)
#        print "r, c, start, vis", r, c, self.start_row, self.fully_visible_rows
        last_row = self.start_row + self.fully_visible_rows - 1
        last_col = self.start_col + self.fully_visible_cols - 1
        
        update = False
        if r < self.start_row:
            # above current view
            update = True
        elif r >= last_row:
            # below last row
            last_scroll_row = self.total_rows - self.fully_visible_rows
            if r >= last_scroll_row:
                r = last_scroll_row
                update = True
            elif r >= self.fully_visible_rows:
                r = r - self.fully_visible_rows + 1
                update = True
        else:
            # row is already visible so don't change row position
            r = self.start_row
        
        if c < self.start_col:
            # left of start column, so set start view to that column
            update = True
        elif c >= last_col:
            c = c - self.fully_visible_cols + 1
            update = True
        else:
            c = self.start_col

        if update:
            self.Scroll(c, r)
        self.Refresh()
    
    def select_addr(self, addr):
        rel_pos = addr - self.start_addr
        self.select_index(rel_pos)

    def on_mouse_wheel(self, evt):
        """Driver to process mouse events.

        This is the main driver to process all mouse events that
        happen on the BitmapScroller.  Once a selector is triggered by
        its event combination, it becomes the active selector and
        further mouse events are directed to its handler.
        """
        if self.end_byte is None:  # end_byte is a proxy for the image being loaded
            return
        
        w = evt.GetWheelRotation()
        if evt.ControlDown():
            if w < 0:
                self.zoom_out()
            elif w > 0:
                self.zoom_in()

        evt.Skip()

    def on_left_up(self, evt):
        self.multi_select_mode = False
        self.select_extend_mode = False
        evt.Skip()

    def set_cursor_pos_from_event(self, evt):
        e = self.editor
        byte, bit, inside = self.event_coords_to_byte(evt)
        if inside:
            self.select_extend_mode = evt.ShiftDown()
            if self.select_extend_mode:
                if byte < e.anchor_start_index:
                    e.anchor_start_index = byte
                elif byte + 1 > e.anchor_start_index:
                    e.anchor_end_index = byte + 1
                e.anchor_initial_start_index, e.anchor_initial_end_index = e.anchor_start_index, e.anchor_end_index
                e.cursor_index = byte
            else:
                e.set_cursor(byte, False)
            wx.CallAfter(e.index_clicked, byte, bit, None)

    def on_left_down(self, evt):
        self.set_cursor_pos_from_event(evt)
        if evt.ControlDown():
            self.multi_select_mode = True
            self.select_extend_mode = False
        elif evt.ShiftDown():
            self.multi_select_mode = False
            self.select_extend_mode = True
        e = self.editor
        byte, bit, inside = self.event_coords_to_byte(evt)
        index1 = byte
        index2 = byte + 1
        if self.select_extend_mode:
            if index1 < e.anchor_start_index:
                e.anchor_start_index = index1
                e.cursor_index = index1
            elif index2 > e.anchor_start_index:
                e.anchor_end_index = index2
                e.cursor_index = index2 - 1
            e.anchor_initial_start_index, e.anchor_initial_end_index = e.anchor_start_index, e.anchor_end_index
            e.select_range(e.anchor_start_index, e.anchor_end_index, add=self.multi_select_mode)
        else:
            e.anchor_initial_start_index, e.anchor_initial_end_index = index1, index2
            e.cursor_index = index1
            e.select_range(index1, index1, add=self.multi_select_mode)
        evt.Skip()

    def on_left_dclick(self, evt):
        self.on_left_down(evt)
 
    def set_selection_from_event(self, evt):
        e = self.editor
        byte, bit, inside = self.event_coords_to_byte(evt)
        if inside:
            index1 = byte
            index2 = byte + 1
            update = False
            if self.select_extend_mode:
                if index1 < e.anchor_initial_start_index:
                    e.select_range(index1, e.anchor_initial_end_index, extend=True)
                    update = True
                else:
                    e.select_range(e.anchor_initial_start_index, index2, extend=True)
                    update = True
            else:
                if e.anchor_start_index <= index1:
                    if index2 != e.anchor_end_index:
                        e.select_range(e.anchor_initial_start_index, index2, extend=self.multi_select_mode)
                        update = True
                else:
                    if index1 != e.anchor_end_index:
                        e.select_range(e.anchor_initial_end_index, index1, extend=self.multi_select_mode)
                        update = True
            if update:
                wx.CallAfter(e.index_clicked, index1, bit, None)
 
    def on_motion(self, evt):
        self.on_motion_update_status(evt)
        if self.editor is not None and evt.LeftIsDown():
            self.set_selection_from_event(evt)
        evt.Skip()

    def on_motion_update_status(self, evt):
        byte, bit, inside = self.event_coords_to_byte(evt)
        if inside:
            label = self.get_label_at_index(byte)
            message = self.get_status_message_at_index(byte, bit)
            self.editor.show_status_message("%s: %s %s" % (self.short_name, label, message))
    
    def get_label_at_index(self, index):
        try:
            label = self.editor.control.table.get_label_at_index(index)
        except AttributeError:
            label = "%x" % index
        return label

    def get_status_message_at_index(self, index, bit):
        comments = self.segment.get_comment(index)
        return "bit=%d  %s" % (bit, comments)

    def on_paint(self, evt):
        self.dbg_call_seq += 1
        log.debug("In on_paint %d" % self.dbg_call_seq)
        self.prepare_image()
        if self.scaled_bmp is not None:
            dc = wx.BufferedPaintDC(self, self.scaled_bmp, wx.BUFFER_CLIENT_AREA)
        evt.Skip()
    
    def on_resize(self, evt):
        if self.is_ready_to_render():
            self.calc_image_size()
    
    def on_popup(self, evt):
        byte, bit, inside = self.event_coords_to_byte(evt)
        actions = self.get_popup_actions()
        style = self.segment.style[byte]
        popup_data = {'index': byte, 'in_selection': style&0x80}
        if actions:
            self.editor.popup_context_menu_from_actions(self, actions, popup_data)
    
    def get_popup_actions(self):
        return self.editor.common_popup_actions()
    
    def on_focus(self, evt):
        log.debug("on_focus!")
    
    def on_focus_lost(self, evt):
        log.debug("on_focus_lost!")
    
    def on_mouse_enter(self, evt):
        evt.Skip()
    
    def on_mouse_leave(self, evt):
        evt.Skip()
    
    def on_char(self, evt):
        log.debug("on_char!")
        evt.Skip()
    
    def process_movement_keys(self, char):
        delta_index = None
        if char == wx.WXK_UP:
            delta_index = -self.bytes_per_row
        elif char == wx.WXK_DOWN:
            delta_index = self.bytes_per_row
        elif char == wx.WXK_LEFT:
            delta_index = -1
        elif char == wx.WXK_RIGHT:
            delta_index = 1
        elif char == wx.WXK_PAGEUP:
            page_size = self.editor.segment.page_size
            if page_size < 0:
                delta_index = -(self.fully_visible_rows * self.bytes_per_row)
            else:
                delta_index = -page_size
        elif char == wx.WXK_PAGEDOWN:
            page_size = self.editor.segment.page_size
            if page_size < 0:
                delta_index = self.fully_visible_rows * self.bytes_per_row
            else:
                delta_index = page_size
        return delta_index
    
    def process_delta_index(self, delta_index):
        e = self.editor
        index = e.set_cursor(e.cursor_index + delta_index, False)
        wx.CallAfter(self.select_index, index)
        wx.CallAfter(e.index_clicked, index, 0, self)
    
    def on_char_hook(self, evt):
        log.debug("on_char_hook! char=%s, key=%s, modifiers=%s" % (evt.GetUniChar(), evt.GetKeyCode(), bin(evt.GetModifiers())))
        mods = evt.GetModifiers()
        char = evt.GetUniChar()
        if char == 0:
            char = evt.GetKeyCode()
        delta_index = self.process_movement_keys(char)
        if delta_index is not None:
            self.process_delta_index(delta_index)
        else:
            evt.Skip()


class BitmapScroller(BitviewScroller):
    short_name = "bitmap"
    
    def update_bytes_per_row(self):
        BitviewScroller.update_bytes_per_row(self)
        self.bytes_per_row = self.editor.machine.bitmap_renderer.validate_bytes_per_row(self.editor.bitmap_width)
    
    def update_zoom(self):
        self.set_zoom(self.editor.bitmap_zoom)
    
    def sync_to_editor(self, e):
        e.bitmap_zoom = self.zoom
        e.bitmap_width = self.bytes_per_row
    
    def event_coords_to_byte(self, evt):
        if self.end_byte is None:  # end_byte is a proxy for the image being loaded
            return 0, 0, False
        
        inside = True

        x, y = self.GetViewStart()
        x = (evt.GetX() // self.zoom) + x
        y = (evt.GetY() // self.zoom) + y
        if x < 0 or x >= self.grid_width or y < 0 or y > (self.start_row + self.visible_rows):
            inside = False
        if x < 0:
            x = 0
        elif x >= self.pixels_per_byte * self.bytes_per_row / self.bitplanes:
            x = self.pixels_per_byte * self.bytes_per_row / self.bitplanes - 1
        xbyte = (x // self.pixels_per_byte) * self.bitplanes
        byte = (self.bytes_per_row * y) + xbyte
        if byte >= self.end_byte:
            inside = False
        bitmask = self.pixels_per_byte - 1
        bit = bitmask - (x & bitmask)
        return byte, bit, inside
    
    def get_image(self, segment=None):
        if segment is None:
            segment = self.segment
        log.debug("get_image: bit image: start=%d, num=%d" % (self.start_row, self.visible_rows))
        sr = self.start_row
        nr = self.visible_rows
        self.start_byte = sr * self.bytes_per_row
        self.end_byte = self.start_byte + (nr * self.bytes_per_row)
        if self.end_byte > len(segment):
            self.end_byte = len(segment)
            count = self.end_byte - self.start_byte
            bytes = np.zeros((nr * self.bytes_per_row), dtype=np.uint8)
            bytes[0:count] = segment[self.start_byte:self.end_byte]
            style = np.zeros((nr * self.bytes_per_row), dtype=np.uint8)
            style[0:count] = segment.style[self.start_byte:self.end_byte]
        else:
            count = self.end_byte - self.start_byte
            bytes = segment[self.start_byte:self.end_byte]
            style = segment.style[self.start_byte:self.end_byte]
        m = self.editor.machine
        array = m.bitmap_renderer.get_image(m, self.bytes_per_row, nr, count, bytes, style)
        sc = self.start_col
        nc = self.visible_cols
        clipped = array[:,sc:sc + nc,:]
        return clipped
    
    def get_popup_actions(self):
        actions = BitviewScroller.get_popup_actions(self)
        actions.extend([None, BitmapWidthAction, BitmapZoomAction])
        return actions


class FontMapScroller(BitviewScroller):
    short_name = "charmap"
    
    def __init__(self, parent, task, bytes_per_row=8, command=None, **kwargs):
        BitviewScroller.__init__(self, parent, task, **kwargs)
        self.zoom = 2
        self.bytes_per_row = bytes_per_row
        self.pixels_per_row = 8
        self.font = None
        self.command_cls = command
        self.inverse = 0
        self.editing = False
        self.blink_index = 0
        self.blink_timer = wx.Timer(self)
        self.Bind(wx.EVT_TIMER, self.update, self.blink_timer)

    def update(self, event):
        self.blink_index = (self.blink_index + 1) % 2
        self.Refresh()
    
    def is_ready_to_render(self):
        return self.font is not None

    def update_bytes_per_row(self):
        BitviewScroller.update_bytes_per_row(self)
        self.bytes_per_row = self.editor.map_width
    
    def update_zoom(self):
        self.set_zoom(self.editor.map_zoom)
    
    def sync_to_editor(self, e):
        e.map_zoom = self.zoom
        e.map_width = self.bytes_per_row
    
    def calc_scale_from_bytes(self):
        self.total_rows = (len(self.segment) + self.bytes_per_row - 1) / self.bytes_per_row
        self.grid_width = int(self.pixels_per_byte * self.bytes_per_row)
        self.grid_height = int(self.pixels_per_row * self.total_rows)
    
    def get_zoom_factors(self):
        zw = self.font.scale_w
        zh = self.font.scale_h
        return zw * self.zoom, zh * self.zoom
    
    def calc_scroll_params(self):
        zw, zh = self.get_zoom_factors()
        self.SetVirtualSize((self.grid_width * zw, self.grid_height * zh))
        self.SetScrollRate(self.pixels_per_byte * zw, self.pixels_per_row * zh)
    
    def calc_image_size(self):
        x, y = self.GetViewStart()
        w, h = self.GetClientSizeTuple()
        self.start_row = y
        self.start_col = y
        
        # For proper buffered paiting, the visible_rows must include the
        # (possibly) partially obscured last row.  fully_visible_rows
        # indicates the number of rows without that last partially obscured
        # row (if it exists).
        zw, zh = self.get_zoom_factors()
        zoom_factor = self.pixels_per_row * zh
        self.fully_visible_rows = h / zoom_factor
        self.visible_rows = (h + zoom_factor - 1) / zoom_factor
        zoom_factor = self.pixels_per_byte * zw
        self.fully_visible_cols = w / zoom_factor
        self.start_col, self.visible_cols = x, (w + zoom_factor - 1) / zoom_factor
        log.debug("fontmap: x, y, w, h, row start, num: %s" % str([x, y, w, h, self.start_row, self.visible_rows, "col start, num:", self.start_col, self.visible_cols]))
    
    def set_font(self):
        self.font = self.editor.machine.antic_font
        self.pixels_per_byte = self.editor.machine.font_renderer.char_bit_width
        self.pixels_per_row = self.editor.machine.font_renderer.char_bit_height
        self.calc_scroll_params()
        if self.font.use_blinking:
            self.blink_timer.Start(267)  # on/off cycle in 1.87 Hz
        else:
            self.blink_timer.Stop()

    def get_image(self, segment=None):
        if segment is None:
            segment = self.segment
        log.debug("get_image: fontmap: start=%d, num=%d" % (self.start_row, self.visible_rows))
        sr = self.start_row
        nr = self.visible_rows
        self.start_byte = sr * self.bytes_per_row
        self.end_byte = self.start_byte + (nr * self.bytes_per_row)
        if self.end_byte > len(segment):
            self.end_byte = len(segment)
            bytes = np.zeros((nr * self.bytes_per_row), dtype=np.uint8)
            bytes[0:self.end_byte - self.start_byte] = segment[sr * self.bytes_per_row:self.end_byte]
            style = np.zeros((nr * self.bytes_per_row), dtype=np.uint8)
            style[0:self.end_byte - self.start_byte] = segment.style[sr * self.bytes_per_row:self.end_byte]
        else:
            bytes = segment[self.start_byte:self.end_byte]
            style = segment.style[self.start_byte:self.end_byte]
        bytes = bytes.reshape((nr, -1))
        style = style.reshape((nr, -1))
        #log.debug("get_image: bytes", bytes)
        
        m = self.editor.machine
        font = m.get_blinking_font(self.blink_index)
        array = m.font_renderer.get_image(m, font, bytes, style, self.start_byte, self.end_byte, self.bytes_per_row, nr, self.start_col, self.visible_cols)
        return array
    
    def draw_overlay(self, array, w, h, zw, zh):
        anchor_start, anchor_end, rc1, rc2 = self.get_highlight_indexes()
        self.show_highlight(array, rc1, rc2, zw, zh)
        r, c = self.byte_to_row_col(self.editor.cursor_index)
        self.show_highlight(array, (r, c), (r + 1, c + 1), zw, zh)
    
    def show_highlight(self, array, rc1, rc2, zw, zh):
        if rc1 is None:
            return
        r1, c1 = rc1
        r2, c2 = rc2
        sr = r1 - self.start_row
        er = r2 - self.start_row
        sc = c1 - self.start_col
        ec = c2 - self.start_col
        x1 = sc * self.pixels_per_byte * zw
        x2 = ec * self.pixels_per_byte * zw - 1
        y1 = sr * self.pixels_per_row * zh
        y2 = er * self.pixels_per_row * zh - 1
        xmax = array.shape[1]
        ymax = array.shape[0]
        c1 = max(x1, 0)
        c2 = min(x2, xmax)
        m = self.editor.machine
        if y1 >= 0 and y1 < ymax and c2 > c1:
            array[y1, c1:c2 + 1] = m.empty_color
        if y2 >= 0 and y2 < ymax and c2 > c1:
            array[y2, c1:c2 + 1] = m.empty_color
        c1 = max(y1, 0)
        c2 = min(y2, ymax)
        if x1 >= 0 and x1 < xmax and c2 > c1:
            array[c1:c2 + 1, x1] = m.empty_color
        if x2 >= 0 and x2 < xmax and c2 > c1:
            array[c1:c2 + 1, x2] = m.empty_color

    def event_coords_to_byte(self, evt):
        """Convert event coordinates to world coordinates.

        Convert the event coordinates to world coordinates by locating
        the offset of the scrolled window's viewport and adjusting the
        event coordinates.
        """
        inside = True

        zw, zh = self.get_zoom_factors()
        x, y = self.GetViewStart()
        x = (evt.GetX() // zw // self.pixels_per_byte) + x
        y = (evt.GetY() // zh // self.pixels_per_row) + y
        if x < 0 or x >= self.bytes_per_row or y < 0 or y > (self.start_row + self.visible_rows):
            inside = False
        byte = (self.bytes_per_row * y) + x
        if byte >= self.end_byte:
            inside = False
        bit = 7 - (y & 7)
        return byte, bit, inside
    
    def get_popup_actions(self):
        actions = BitviewScroller.get_popup_actions(self)
        actions.extend([None, FontMappingWidthAction, FontMappingZoomAction, None])
        actions.extend(self.task.get_font_mapping_actions(self.task))
        return actions
    
    def get_status_message(self):
        if not self.editing:
            message = "Double click or press F2 to begin editing text"
        else:
            if self.inverse:
                message = "Editing text (INVERSE MODE): Press F1 for normal characters"
            else:
                message = "Editing text: Press F1 for inverse"
            message += " (press F2 to stop editing or click outside the character map window)"
        return message

    def get_status_message_at_index(self, index, bit):
        comments = self.segment.get_comment(index)
        return "%s %s" % (comments, self.get_status_message())

    def set_status_message(self):
        e = self.editor
        if e is None:
            return
        e.task.status_bar.message = self.get_status_message()
    
    def on_focus(self, evt):
        log.debug("on_focus!")
        self.inverse = 0
        self.pending_esc = False
        self.set_status_message()
        self.editing = False
        
    def on_focus_lost(self, evt):
        log.debug("on_focus_lost!")
        e = self.editor
        if e is not None:
            e.task.status_bar.message = ""
        self.editing = False

    def on_left_dclick(self, evt):
        e = self.editor
        byte, bit, inside = self.event_coords_to_byte(evt)
        if inside:
            e.set_cursor(byte, False)
            wx.CallAfter(e.index_clicked, byte, bit, None)
            self.editing = True
            self.set_status_message()
        evt.Skip()
    
    def on_char(self, evt):
        log.debug("on_char! char=%s, key=%s, shift=%s, ctrl=%s, cmd=%s" % (evt.GetUniChar(), evt.GetRawKeyCode(), evt.ShiftDown(), evt.ControlDown(), evt.CmdDown()))
        if not self.editing:
            evt.Skip()
            return
        char = evt.GetUniChar()
        if char > 0:
            self.change_byte(char | self.inverse)
    
    def change_byte(self, value):
        e = self.editor
        if e is None:
            return
        cmd_cls = self.command_cls
        if cmd_cls is None:
            return
        if e.can_copy:
            index = e.anchor_start_index
        else:
            index = e.cursor_index
        cmd = cmd_cls(e.segment, index, index+1, value, True)
        e.process_command(cmd)
    
    def on_char_hook(self, evt):
        log.debug("on_char_hook! char=%s, key=%s, modifiers=%s" % (evt.GetUniChar(), evt.GetKeyCode(), bin(evt.GetModifiers())))
        mods = evt.GetModifiers()
        char = evt.GetUniChar()
        if char == 0:
            char = evt.GetKeyCode()
        byte = None
        delta_index = None
        if not self.editing:
            if char == wx.WXK_F2:
                self.editing = True
                self.set_status_message()
            else:
                delta_index = self.process_movement_keys(char)
        else:
            if mods == wx.MOD_RAW_CONTROL:
                if char == 44:  # Ctrl-, prints ATASCII 0 (heart)
                    byte = 0 + self.inverse
                elif char >= 65 and char <= 90:  # Ctrl-[A-Z] prints ATASCII chars 1-26
                    byte = char - 64 + self.inverse
                elif char == 46:  # Ctrl-. prints ATASCII 96 (diamond)
                    byte = 96 + self.inverse
                elif char == 59:  # Ctrl-; prints ATASCII 123 (spade)
                    byte = 123 + self.inverse
                elif char == wx.WXK_TAB:
                    byte = 158
                elif char == 50:  # Ctrl-2 prints ATASCII 253 (buzzer)
                    byte = 253
                elif char == wx.WXK_INSERT:
                    byte = 255
            elif mods == wx.MOD_SHIFT:
                if char == wx.WXK_BACK:
                    byte = 156
                elif char == wx.WXK_INSERT:
                    byte = 157
                elif char == wx.WXK_TAB:
                    byte = 159
            elif char == wx.WXK_HOME:
                byte = 125
            elif char == wx.WXK_BACK:
                byte = 126
            elif char == wx.WXK_TAB:
                byte = 127
            elif char == wx.WXK_RETURN:
                byte = 155
            elif char == wx.WXK_DELETE:
                byte = 254
            elif char == wx.WXK_INSERT:
                byte = 255
            
            elif char == wx.WXK_F1:
                self.inverse = (self.inverse + 0x80) & 0x80
                self.set_status_message()
            elif char == wx.WXK_F2:
                self.editing = False
                self.set_status_message()
            
            elif self.pending_esc:
                if char == wx.WXK_ESCAPE:
                    byte = 27
                elif char == wx.WXK_UP:
                    byte = 28
                elif char == wx.WXK_DOWN:
                    byte = 29
                elif char == wx.WXK_LEFT:
                    byte = 30
                elif char == wx.WXK_RIGHT:
                    byte = 31
            
            elif char == wx.WXK_ESCAPE:
                self.pending_esc = True
        
            else:
                delta_index = self.process_movement_keys(char)
        
        if byte is not None:
            self.change_byte(byte)
            self.pending_esc = False
        elif delta_index is not None:
            e = self.editor
            index = e.set_cursor(e.cursor_index + delta_index, False)
            wx.CallAfter(self.select_index, index)
            wx.CallAfter(e.index_clicked, index, 0, self)
        else:
            evt.Skip()


class CharacterSetViewer(FontMapScroller):
    def __init__(self, parent, task, bytes_per_row=16, command=None, **kwargs):
        FontMapScroller.__init__(self, parent, task, bytes_per_row, command, **kwargs)
        self.segment = DefaultSegment(SegmentData(np.arange(256, dtype=np.uint8), np.zeros(256, dtype=np.uint8)), 0)
        self.start_addr = 0
        self.selected_char = -1
    
    def set_selected_char(self, index):
        self.selected_char = index
        e = self.editor
        if e is not None:
            e.set_current_draw_pattern(self.selected_char, self)
    
    def clear_tile_selection(self):
        self.selected_char = -1
        self.Refresh()
    
    def recalc_view(self):
        editor = self.task.active_editor
        if editor is not None:
            self.editor = editor
            self.set_colors()
            self.set_font()
            self.set_scale()

    def on_left_down(self, evt):
        e = self.editor
        byte, bit, inside = self.event_coords_to_byte(evt)
        if inside:
            self.set_selected_char(byte)
            wx.CallAfter(self.Refresh)
        evt.Skip()
 
    def on_left_dclick(self, evt):
        e = self.editor
        byte, bit, inside = self.event_coords_to_byte(evt)
        if inside:
            self.set_selected_char(byte)
            wx.CallAfter(self.Refresh)
        evt.Skip()
    
    def on_motion(self, evt):
        self.on_motion_update_status(evt)
        e = self.editor
        if e is not None and evt.LeftIsDown():
            byte, bit, inside = self.event_coords_to_byte(evt)
            if inside:
                pass
        evt.Skip()
    
    def set_status_message(self):
        return
    
    def get_popup_actions(self):
        return []
    
    def get_highlight_indexes(self):
        if self.selected_char < 0:
            return 0, 0, None, None
        return self.selected_char, self.selected_char + 1, None, None
    
    def process_delta_index(self, delta_index):
        _, byte = divmod(self.selected_char + delta_index, 256)
        self.set_selected_char(byte)
        self.Refresh()
    
    def on_char_hook(self, evt):
        log.debug("on_char_hook! char=%s, key=%s, modifiers=%s" % (evt.GetUniChar(), evt.GetKeyCode(), bin(evt.GetModifiers())))
        mods = evt.GetModifiers()
        char = evt.GetUniChar()
        if char == 0:
            char = evt.GetKeyCode()
        delta_index = self.process_movement_keys(char)
        if delta_index is not None:
            self.process_delta_index(delta_index)
        else:
            evt.Skip()


class MemoryMapScroller(BitviewScroller):
    def __init__(self, parent, task, **kwargs):
        BitviewScroller.__init__(self, parent, task, **kwargs)
        self.bytes_per_row = 256
        self.zoom = 2
    
    def DoGetBestSize(self):
        """ Base class virtual method for sizer use to get the best size
        """
        width = self.bytes_per_row * self.zoom
        height = -1
        
        vw, vh = self.GetVirtualSizeTuple()
        ww, wh = self.GetClientSizeTuple()
        if wh < vh:
            # Scrollbar is present!
            width += wx.SystemSettings.GetMetric(wx.SYS_VSCROLL_X) + 1
        best = wx.Size(width, height)

        # Cache the best size so it doesn't need to be calculated again,
        # at least until some properties of the window change
        self.CacheBestSize(best)

        return best
    
    def calc_scale_from_bytes(self):
        self.total_rows = (len(self.segment) + self.bytes_per_row - 1) / self.bytes_per_row
        self.grid_width = int(self.bytes_per_row)
        self.grid_height = int(self.total_rows)
    
    def calc_scroll_params(self):
        z = self.zoom
        self.SetVirtualSize((self.grid_width * z, self.grid_height * z))
        self.SetScrollRate(z, z)
    
    def calc_image_size(self):
        x, y = self.GetViewStart()
        w, h = self.GetClientSizeTuple()
        self.start_row = y
        self.start_col = x
        
        # For proper buffered paiting, the visible_rows must include the
        # (possibly) partially obscured last row.  fully_visible_rows
        # indicates the number of rows without that last partially obscured
        # row (if it exists).
        z = self.zoom
        self.fully_visible_rows = h / z
        self.fully_visible_cols = w / z
        self.visible_rows = (h + z - 1) / z
        self.start_col, self.visible_cols = x, (w + z - 1) / z
        log.debug("memory map: x, y, w, h, row start, num: %s" % str([x, y, w, h, self.start_row, self.visible_rows, "col start, num:", self.start_col, self.visible_cols]))

    def get_image(self, segment=None):
        if segment is None:
            segment = self.segment
        log.debug("get_image: memory map: start=%d, num=%d" % (self.start_row, self.visible_rows))
        t0 = time.clock()
        sr = self.start_row
        nr = self.visible_rows
        self.start_byte = sr * self.bytes_per_row
        self.end_byte = self.start_byte + (nr * self.bytes_per_row)
        if self.end_byte > len(segment):
            self.end_byte = len(segment)
            bytes = np.zeros((nr * self.bytes_per_row), dtype=np.uint8)
            bytes[0:self.end_byte - self.start_byte] = segment[sr * self.bytes_per_row:self.end_byte]
            style = np.zeros((nr * self.bytes_per_row), dtype=np.uint8)
            style[0:self.end_byte - self.start_byte] = segment.style[sr * self.bytes_per_row:self.end_byte]
        else:
            bytes = segment[self.start_byte:self.end_byte]
            style = segment.style[self.start_byte:self.end_byte]
        bytes = bytes.reshape((nr, -1))
        style = style.reshape((nr, -1))
        #log.debug("get_image: bytes", bytes)
        
        m = self.editor.machine
        array = m.page_renderer.get_image(m, bytes, style, self.start_byte, self.end_byte, self.bytes_per_row, nr, self.start_col, self.visible_cols)
        log.debug(array.shape)
        t = time.clock()
        log.debug("get_image: time %f" % (t - t0))
        return array

    def event_coords_to_byte(self, evt):
        """Convert event coordinates to world coordinates.

        Convert the event coordinates to world coordinates by locating
        the offset of the scrolled window's viewport and adjusting the
        event coordinates.
        """
        inside = True

        z = self.zoom
        x, y = self.GetViewStart()
        x = (evt.GetX() // z) + x
        y = (evt.GetY() // z) + y
        if x < 0 or x >= self.bytes_per_row or y < 0 or y > (self.start_row + self.visible_rows):
            inside = False
        byte = (self.bytes_per_row * y) + x
        if byte >= self.end_byte:
            inside = False
        return byte, 0, inside


if __name__ == '__main__':
    app   = wx.PySimpleApp()
    frame = wx.Frame(None, -1, title='Test', size=(500,500))
    frame.CreateStatusBar()
    
    panel = BitviewScroller(frame)
    bytes = np.arange(256, dtype=np.uint8)
    panel.set_data(bytes)

    sizer = wx.BoxSizer(wx.VERTICAL)
    sizer.Add(panel,  1, wx.EXPAND | wx.ALL, 5)
    
    def buttonHandler(evt):
        id = evt.GetId()
        if id == 100:
            panel.zoom_in()
        elif id == 101:
            panel.zoom_out()
        elif id == 200:
            wildcard="*"
            dlg = wx.FileDialog(
                frame, message="Open File",
                defaultFile="", wildcard=wildcard, style=wx.OPEN)

            # Show the dialog and retrieve the user response. If it is the
            # OK response, process the data.
            if dlg.ShowModal() == wx.ID_OK:
                # This returns a Python list of files that were selected.
                paths = dlg.GetPaths()

                for path in paths:
                    dprint("open file %s:" % path)
                    fh = open(path, 'rb')
                    img = wx.EmptyImage()
                    if img.LoadStream(fh):
                        panel.setImage(img)
                    else:
                        dprint("Invalid image: %s" % path)
            # Destroy the dialog. Don't do this until you are done with it!
            # BAD things can happen otherwise!
            dlg.Destroy()
        elif id == 201:
            pass
        elif id == 202:
            panel.copy_to_clipboard()
    buttonsizer = wx.BoxSizer(wx.HORIZONTAL)
    sizer.Add(buttonsizer, 0, wx.EXPAND | wx.ALL, 5)
    button = wx.Button(frame, 100, 'Zoom In')
    frame.Bind(wx.EVT_BUTTON, buttonHandler, button)
    buttonsizer.Add(button, 0, wx.EXPAND, 0)
    button = wx.Button(frame, 101, 'Zoom Out')
    frame.Bind(wx.EVT_BUTTON, buttonHandler, button)
    buttonsizer.Add(button, 0, wx.EXPAND, 0)
    button = wx.Button(frame, 200, 'Load')
    frame.Bind(wx.EVT_BUTTON, buttonHandler, button)
    buttonsizer.Add(button, 0, wx.EXPAND, 0)
    button = wx.Button(frame, 202, 'Copy to Clipboard')
    frame.Bind(wx.EVT_BUTTON, buttonHandler, button)
    buttonsizer.Add(button, 0, wx.EXPAND, 0)

    frame.SetAutoLayout(1)
    frame.SetSizer(sizer)
    frame.Show(1)
    app.MainLoop()



# Standard library imports.
import sys
import os
import cPickle as pickle

# Major package imports.
import wx
import numpy as np
from atrcopy import SegmentData, DefaultSegment, selected_bit_mask, comment_bit_mask, data_bit_mask, match_bit_mask

# Enthought library imports.
from traits.api import on_trait_change, Any, Bool, Int, Str, List, Event, Enum, Instance, File, Unicode, Property, provides
from pyface.api import YES, NO

# Local imports.
from omnivore.tasks.hex_edit.hex_editor import HexEditor
from omnivore.tasks.bitmap_edit.bitmap_editor import MainBitmapScroller, BitmapEditor
from omnivore.framework.document import Document
from omnivore.arch.machine import Machine, predefined
from omnivore.arch.colors import powerup_colors
from omnivore.arch.antic_renderers import BaseRenderer
from omnivore.utils.wx.bitviewscroller import BitmapScroller
from omnivore.utils.jumpman import *
from omnivore.templates import get_template

from commands import *
from actions import *
from mouse_modes import *

import logging
log = logging.getLogger(__name__)


class JumpmanLevelView(MainBitmapScroller):
    default_mouse_handler = JumpmanSelectMode

    def __init__(self, *args, **kwargs):
        MainBitmapScroller.__init__(self, *args, **kwargs)
        self.level_builder = None
        self.trigger_root = None
        self.cached_screen = None
        self.valid_level = False
        self.force_refresh = True
        self.screen_state = None
        self.trigger_state = None
    
    def is_ready_to_render(self):
        return self.editor is not None and self.level_builder is not None and self.segment is not None
    
    def on_resize(self, evt):
        # Automatically resize image to a best fit when resized
        if self.is_ready_to_render():
            self.update_zoom()
            self.calc_image_size()
            self.set_scale()

    def update_zoom(self):
        # force zoom so that entire screen fits in window
        w, h = self.GetClientSizeTuple()
        sw = w / 160
        sh = h / 88
        zoom = min(sh, sw)
        self.set_zoom(zoom)

    def set_mouse_mode(self, handler):
        if hasattr(self, 'mouse_mode'):
            self.mouse_mode.cleanup()
        self.release_mouse()
        self.mouse_mode = handler(self)

    def get_segment(self, editor):
        self.level_builder = JumpmanLevelBuilder(editor.document.user_segments)
        self.trigger_root = None
        self.pick_buffer = editor.pick_buffer
        self.force_refresh = True
        self.generate_display_objects()
        return editor.screen

    def generate_display_objects(self, resync=False):
        try:
            source, level_addr, harvest_addr = self.editor.get_level_addrs()
            index = level_addr - source.start_addr
            self.level_builder.parse_level_data(source, level_addr, harvest_addr)
            self.force_refresh = True
            self.valid_level = True
            if self.trigger_root is not None:
                self.trigger_root = self.level_builder.find_equivalent_peanut(self.trigger_root)
                self.editor.trigger_list.recalc_view()
            if resync:
                self.mouse_mode.resync_objects()
        except RuntimeError:
            self.valid_level = False

    def clear_screen(self):
        self.editor.clear_playfield()
        self.pick_buffer[:] = -1

    def set_trigger_root(self, root):
        if root is not None:
            root = self.level_builder.find_equivalent_peanut(root)
        self.trigger_root = root
        self.force_refresh = True
        self.trigger_state = None

    def get_screen_state(self):
        if self.trigger_state is not None:
            return self.trigger_state
        return self.screen_state

    def redraw_current(self, screen, overlay_objects=[]):
        e = self.editor
        e.clear_playfield(screen)
        self.pick_buffer[:] = -1
        self.level_builder.set_harvest_offset(self.mouse_mode.get_harvest_offset())
        main_state = self.level_builder.draw_objects(screen, None, e.segment, highlight=overlay_objects, pick_buffer=self.pick_buffer)
        log.debug("draw objects: %s" % self.level_builder.objects)
        if main_state.missing_object_codes:
            log.error("missing draw codes: %s" % (sorted(main_state.missing_object_codes)))
        if self.trigger_root is not None:
            self.level_builder.fade_screen(screen)
            root = [self.trigger_root]
            self.level_builder.draw_objects(screen, root, e.segment, highlight=root, pick_buffer=self.pick_buffer)
            # change highlight to comment color for selected trigger peanut so
            # you don't get confused with any objects actually selected
            old_highlight = np.where(screen.style == selected_bit_mask)
            screen.style[old_highlight] |= comment_bit_mask
            screen.style[:] &= (0xff ^ (match_bit_mask|selected_bit_mask))

            # replace screen state so that the only pickable objects are
            # those in the triggered layer
            self.pick_buffer[:] = -1
            trigger_state = self.level_builder.draw_objects(screen, self.trigger_root.trigger_painting, e.segment, highlight=overlay_objects, pick_buffer=self.pick_buffer)
            active_state = trigger_state
        else:
            active_state = main_state
            trigger_state = None
        return main_state, trigger_state, active_state

    def compute_image(self, force=False):
        if force:
            self.force_refresh = True
        if self.force_refresh:
            self.screen_state, self.trigger_state, _ = self.redraw_current(self.segment)
            self.cached_screen = self.segment[:].copy()
            self.force_refresh = False
            self.editor.update_harvest_state()
        else:
            self.segment[:] = self.cached_screen

    def bad_image(self):
        self.segment[:] = 0
        self.segment.style[:] = 0
        self.force_refresh = True
        s = self.segment.style.reshape((self.editor.antic_lines, -1))
        s[::2,::2] = comment_bit_mask
        s[1::2,1::2] = comment_bit_mask

    def get_rendered_image(self, segment=None):
        return MainBitmapScroller.get_image(self, segment)

    def get_image(self):
        if self.editor.valid_jumpman_segment:
            override = self.mouse_mode.get_image_override()
            if override is not None:
                return override
        if self.valid_level:
            self.compute_image()
        else:
            self.bad_image()
            bitimage = MainBitmapScroller.get_image(self)
            return bitimage
        self.mouse_mode.draw_extra_objects(self.level_builder, self.segment, self.editor.segment)
        bitimage = MainBitmapScroller.get_image(self)
        self.mouse_mode.draw_overlay(bitimage)
        return bitimage

    def get_save_location(self):
        if self.trigger_root is not None:
            equiv = self.level_builder.find_equivalent_peanut(self.trigger_root)
            parent = equiv.trigger_painting
        else:
            parent = None
        return parent

    def delete_objects(self, objects):
        save_location = self.get_save_location()
        self.level_builder.delete_objects(objects, save_location)
        self.save_changes()

    def save_objects(self, objects, command_cls=CreateObjectCommand):
        save_location = self.get_save_location()
        self.level_builder.add_objects(objects, save_location)
        if self.trigger_root is not None:
            print save_location, id(save_location)
            print self.trigger_root, id(self.trigger_root), id(self.trigger_root.trigger_painting)
        self.save_changes(command_cls)

    def save_changes(self, command_cls=MoveObjectCommand):
        source, level_addr, old_harvest_addr = self.editor.get_level_addrs()
        level_data, harvest_addr, ropeladder_data, num_peanuts = self.level_builder.create_level_definition(level_addr, source[0x46], source[0x47])
        index = level_addr - source.start_addr
        ranges = [(0x18,0x2a), (0x3e,0x3f), (0x4e,0x50), (index,index + len(level_data))]
        pdata = np.empty([1], dtype=np.uint8)
        if self.editor.peanut_harvest_diff < 0:
            pdata[0] = num_peanuts
        else:
            pdata[0] = max(0, num_peanuts - self.editor.peanut_harvest_diff)
        hdata = np.empty([2], dtype=np.uint8)
        hdata.view(dtype="<u2")[0] = harvest_addr
        data = np.hstack([ropeladder_data, pdata, hdata, level_data])
        cmd = command_cls(source, ranges, data)
        self.editor.process_command(cmd)
    
    # Segment saver interface for menu item display
    export_data_name = "Jumpman Level Tester ATR"
    export_extensions = [".atr"]
    
    def encode_data(self, segment):
        """Segment saver interface: take a segment and produce a byte
        representation to save to disk.
        """
        image = get_template("Jumpman Level")
        if image is None:
            raise RuntimeError("Can't find Jumpman Level template file")
        raw = np.fromstring(image, dtype=np.uint8)
        raw[0x0196:0x0996] = segment[:]
        return raw.tostring()


class JumpmanPlayfieldRenderer(BaseRenderer):
    """ Custom renderer instead of Antic Mode D renderer. Need to display
    highlighting on a per-pixel level which isn't possible with the Mode D
    renderer because the styling info is applied at the byte level and there
    are 4 pixels per byte in Mode D.

    So this renderer is one byte per pixel, but only uses the first 16 colors.
    It is mapped to the ANTIC color register order, so the first 4 colors are
    player colors, then the 5 playfield colors. A blank screen corresponds to
    the index value of 8, so the last playfield color.
    """
    name = "Jumpman 1 Byte Per Pixel"
    pixels_per_byte = 1
    bitplanes = 1

    def get_image(self, m, bytes_per_row, nr, count, bytes, style):
        normal = style == 0
        highlight = (style & selected_bit_mask) == selected_bit_mask
        comment = (style & comment_bit_mask) == comment_bit_mask
        data = (style & data_bit_mask) == data_bit_mask
        match = (style & match_bit_mask) == match_bit_mask
        
        color_registers, h_colors, m_colors, c_colors, d_colors = self.get_colors(m, range(32))
        bitimage = np.empty((nr * bytes_per_row, 3), dtype=np.uint8)
        for i in range(32):
            color_is_set = (bytes == i)
            bitimage[color_is_set & normal] = color_registers[i]
            bitimage[color_is_set & comment] = c_colors[i]
            bitimage[color_is_set & match] = m_colors[i]
            bitimage[color_is_set & data] = d_colors[i]
            bitimage[color_is_set & highlight] = h_colors[i]
        bitimage[count:,:] = m.empty_color
        return bitimage.reshape((nr, bytes_per_row, 3))


class JumpmanEditor(BitmapEditor):
    """ The toolkit specific implementation of a HexEditor.  See the
    IHexEditor interface for the API documentation.
    """
    valid_jumpman_segment = Bool

    peanut_harvest_diff = Int(-1)

    num_ladders = Int(-1)

    num_downropes = Int(-1)

    num_peanuts = Int(-1)

    can_select_objects = Bool(False)

    can_erase_objects = Bool(False)

    ##### class attributes
    
    valid_mouse_modes = [AnticDSelectMode, DrawGirderMode, DrawLadderMode, DrawUpRopeMode, DrawDownRopeMode, DrawPeanutMode, EraseGirderMode, EraseLadderMode, EraseRopeMode, JumpmanRespawnMode]
    
    ##### Default traits
    
    def _machine_default(self):
        return Machine(name="Jumpman", bitmap_renderer=JumpmanPlayfieldRenderer(), mime_prefix="application/vnd.atari8bit")

    def _map_width_default(self):
        return 40 * 4
    
    def _draw_pattern_default(self):
        return [0]
    
    def _valid_jumpman_segment_default(self):
        return False

    def _mouse_mode_default(self):
        return AnticDSelectMode

    ###########################################################################
    # 'FrameworkEditor' interface.
    ###########################################################################
    
    # Segment saver interface for menu item display
    export_data_name = "Jumpman Level Tester ATR"
    export_extensions = [".atr"]

    def is_valid_for_save(self):
        all_ok = True
        if not self.bitmap.level_builder.harvest_ok:
            reason = self.bitmap.level_builder.harvest_reason()
            answer = self.task.confirm("%s\n\nSave Anyway?" % reason, "Bad Peanut Grid")
            all_ok = (answer == YES)
        return all_ok

    def made_current_active_editor(self):
        self.update_mouse_mode(AnticDSelectMode)
        self.refresh_toolbar_state()

    def process_extra_metadata(self, doc, e):
        # ignore bitmap renderer in restore because we always want to use the
        # JumpmanPlayfieldRenderer in Jumpman level edit mode
        if 'bitmap_renderer' in e:
            del e['bitmap_renderer']
        BitmapEditor.process_extra_metadata(self, doc, e)
    
    @on_trait_change('machine.bitmap_shape_change_event')
    def update_bitmap_shape(self):
        self.hex_edit.recalc_view()
        self.bitmap.recalc_view()
    
    @on_trait_change('machine.bitmap_color_change_event')
    def update_bitmap_colors(self):
        try:
            self.bitmap.compute_image(True)
        except RuntimeError:
            pass
        self.bitmap.mouse_mode.resync_objects()
    
    @on_trait_change('machine.font_change_event')
    def update_fonts(self):
        pass
    
    @on_trait_change('machine.disassembler_change_event')
    def update_disassembler(self):
        pass

    def rebuild_display_objects(self):
        print "Rebuilding!"
        self.bitmap.generate_display_objects(True)
    
    def reconfigure_panes(self):
        self.hex_edit.recalc_view()
        self.bitmap.recalc_view()
        self.level_data.recalc_view()
        self.trigger_list.recalc_view()
    
    def refresh_panes(self):
        self.hex_edit.refresh_view()
        p = self.get_level_colors()
        if p != self.machine.antic_color_registers:
            self.machine.update_colors(p)
        self.bitmap.refresh_view()
        self.trigger_list.refresh_view()
        # level_data is refreshed with call to update_harvest_state
    
    def rebuild_document_properties(self):
        self.update_mouse_mode(AnticDSelectMode)

    def check_valid_segment(self, segment=None):
        if segment is None:
            segment = self.segment
        # 283f is always 4c (JMP) because it and the next two bytes are a jump target from the game loop
        # 2848: always 20 (i.e. JSR)
        # 284b: always 60 (i.e. RTS)
        # 284c: always FF (target for harvest table if no action to be taken)
        if len(segment) >= 0x800 and segment[0x3f] == 0x4c and segment[0x48] == 0x20 and segment[0x4b] == 0x60 and segment[0x4c] == 0xff:
            # check for sane level definition table
            index = segment[0x38]*256 + segment[0x37] - segment.start_addr
            return index >=0 and index < len(segment)
        return False

    def get_level_addrs(self):
        if not self.valid_jumpman_segment:
            raise RuntimeError
        source = self.segment
        start = source.start_addr
        level_addr = source[0x37] + source[0x38]*256
        harvest_addr = source[0x4e] + source[0x4f]*256
        log.debug("level def table: %x, harvest table: %x" % (level_addr, harvest_addr))
        last = source.start_addr + len(source)
        if level_addr > start and harvest_addr > start and level_addr < last and harvest_addr < last:
            return source, level_addr, harvest_addr
        raise RuntimeError

    def find_first_valid_segment_index(self):
        # Find list of matches meeting minimum criteria
        possible = []
        for i, segment in enumerate(self.document.segments):
            if self.check_valid_segment(segment):
                possible.append(i)
                if len(segment) == 0x800:
                    return i
        if possible:
            return possible[0]
        return 0
    
    def init_view_properties(self):
        self.find_segment()

    def copy_view_properties(self, old_editor):
        segment = old_editor.segment
        if self.check_valid_segment(segment):
            if len(segment) != 0x800:
                segment = None
        self.find_segment(segment=segment)
    
    def view_segment_set_width(self, segment):
        self.valid_jumpman_segment = self.check_valid_segment(segment)
        self.bitmap_width = 40 * 4
        self.machine.update_colors(self.get_level_colors(segment))
        self.update_mouse_mode()
        self.peanut_harvest_diff = -1
    
    def update_harvest_state(self):
        if not self.valid_jumpman_segment:
            return
        harvest_state = self.bitmap.level_builder.get_harvest_state()
        self.num_ladders = len(harvest_state.ladder_positions)
        self.num_downropes = len(harvest_state.downrope_positions)
        self.num_peanuts = len(harvest_state.peanuts)
        
        # FIXME: force redraw of level data here because it depends on the
        # level builder objects so it can count the number of items
        self.level_data.refresh_view()
        self.trigger_list.refresh_view()

    def get_level_colors(self, segment=None):
        if segment is None:
            segment = self.segment
        if self.check_valid_segment(segment):
            colors = segment[0x2a:0x33].copy()
            # on some levels, the bombs are set to color 0 because they are
            # cycled to produce a glowing effect, but that's not visible here
            # so we force it to be bright white
            fg = colors[4:8]
            fg[fg == 0] = 15
        else:
            colors = powerup_colors()
        return list(colors)
    
    def update_mouse_mode(self, mouse_handler=None):
        BitmapEditor.update_mouse_mode(self, mouse_handler)
        self.can_select_objects = self.bitmap.mouse_mode.can_paste
        self.bitmap.refresh_view()
    
    def set_current_draw_pattern(self, pattern, control):
        try:
            iter(pattern)
        except TypeError:
            self.draw_pattern = [pattern]
        else:
            self.draw_pattern = pattern
        if control != self.tile_map:
            self.tile_map.clear_tile_selection()
        if control != self.character_set:
            self.character_set.clear_tile_selection()
    
    def highlight_selected_ranges(self):
        HexEditor.highlight_selected_ranges(self)

    def mark_index_range_changed(self, index_range):
        pass
    
    def perform_idle(self):
        mouse_mode = self.bitmap.mouse_mode
        self.can_copy = self.can_cut = mouse_mode.can_paste and bool(mouse_mode.objects)
    
    def process_paste_data_object(self, data_obj, cmd_cls=None):
        # Don't use bitmap editor's paste, we want it to paste in hex
        return HexEditor.process_paste_data_object(self, data_obj, cmd_cls)
    
    def create_clipboard_data_object(self):
        # Don't use bitmap editor's clipboard, we want hex bytes
        return HexEditor.create_clipboard_data_object(self)
    
    def get_extra_segment_savers(self, segment):
        savers = []
        savers.append(self.bitmap)
        return savers

    def get_playfield(self):
        data = np.empty(40 * 4 * self.antic_lines, dtype=np.uint8)
        return data

    def get_playfield_segment(self, playfield=None):
        if playfield is None:
            playfield = self.get_playfield()
        r = SegmentData(playfield)
        return DefaultSegment(r, 0x7000)

    def clear_playfield(self, playfield=None):
        if playfield is None:
            playfield = self.screen
        playfield[:] = 8  # background is the 9th ANTIC color register
        playfield.style[:] = 0
    
    def undo_post_hook(self):
        self.rebuild_display_objects()
        self.update_mouse_mode()
    
    def redo_post_hook(self):
        self.rebuild_display_objects()
        self.update_mouse_mode()

    ##### Copy/Paste support

    def cut(self):
        """ Cuts the current selection to the clipboard
        """
        if self.copy():
            self.bitmap.mouse_mode.delete_objects()

    def process_paste_data_object(self, data_obj, cmd_cls=None):
        mouse_mode = self.bitmap.mouse_mode
        if mouse_mode.can_paste:
            value = data_obj.GetData()
            fmt = data_obj.GetPreferredFormat()
            if fmt.GetId() == "jumpman,objects":
                objects = pickle.loads(value)
                for o in objects:
                    # offset slightly so the pasted objects are seen
                    o.x += 1
                    o.y += 1
                mouse_mode.objects = objects
                self.bitmap.save_objects(objects)
    
    supported_clipboard_data_objects = [
        wx.CustomDataObject("jumpman,objects"),
        ]
    
    def create_clipboard_data_object(self):
        mouse_mode = self.bitmap.mouse_mode
        if mouse_mode.can_paste and mouse_mode.objects:
            data = mouse_mode.objects
            data_obj = wx.CustomDataObject("jumpman,objects")
            data_obj.SetData(pickle.dumps(data))
            return data_obj
        return None

    def select_all(self, refresh=True):
        """ Selects the entire document
        """
        mouse_mode = self.bitmap.mouse_mode
        if mouse_mode.can_paste:
            first = True
            for obj in self.bitmap.level_builder.objects:
                mouse_mode.add_to_selection(obj, not first)
                first = False
            if refresh:
                self.refresh_panes()

    def select_none(self, refresh=True):
        """ Clears any selection in the document
        """
        mouse_mode = self.bitmap.mouse_mode
        if mouse_mode.can_paste:
            mouse_mode.objects = []
            if refresh:
                self.refresh_panes()

    def select_invert(self, refresh=True):
        """ Selects the entire document
        """
        mouse_mode = self.bitmap.mouse_mode
        if mouse_mode.can_paste:
            first = True
            current = set(mouse_mode.objects)
            for obj in self.bitmap.level_builder.objects:
                if obj not in current:
                    mouse_mode.add_to_selection(obj, not first)
                    first = False
            if refresh:
                self.refresh_panes()

    def set_trigger_view(self, trigger_root):
        mouse_mode = self.bitmap.mouse_mode
        if mouse_mode.can_paste:
            mouse_mode.objects = []
        self.bitmap.set_trigger_root(trigger_root)
        self.can_erase_objects = trigger_root is not None
        self.refresh_panes()
        self.refresh_toolbar_state()

    ###########################################################################
    # Trait handlers.
    ###########################################################################


    ###########################################################################
    # Private interface.
    ###########################################################################

    def _create_control(self, parent):
        """ Creates the toolkit-specific control for the widget. """

        # Base-class constructor.
        self.bitmap = JumpmanLevelView(parent, self.task)

        self.antic_lines = 88
        self.screen = self.get_playfield_segment()
        self.pick_buffer = np.zeros((160 * self.antic_lines), dtype=np.int32)

        ##########################################
        # Events.
        ##########################################

        # Get related controls
        self.segment_list = self.window.get_dock_pane('jumpman.segments').control
        self.undo_history = self.window.get_dock_pane('jumpman.undo').control
        self.hex_edit = self.window.get_dock_pane('jumpman.hex').control
        self.level_data = self.window.get_dock_pane('jumpman.level_data').control
        self.trigger_list = self.window.get_dock_pane('jumpman.triggers').control

        # Load the editor's contents.
        self.load()

        return self.bitmap

    #### wx event handlers ####################################################
    
    def index_clicked(self, index, bit, control):
        self.cursor_index = index
        if control != self.hex_edit:
            self.hex_edit.select_index(index)



#-------------------------------------------------------------------------------
#
#  Copyright (c) 2007, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   07/09/2007
#
#-------------------------------------------------------------------------------

""" Defines a ThemedWindow base class for creating themed windows.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

import wx

from traits.api \
    import HasPrivateTraits, HasTraits, Instance, Str, Enum, Any, Bool

from traitsui.ui_traits \
    import ATheme

from helper \
    import init_wx_handlers, BufferDC

#-------------------------------------------------------------------------------
#  'ThemedWindow' class:
#-------------------------------------------------------------------------------

class ThemedWindow ( HasPrivateTraits ):

    #-- Public Traits ----------------------------------------------------------

    # The theme associated with this window:
    theme = ATheme

    # The default alignment to use:
    default_alignment = Enum( 'left', 'center', 'right' )

    # The current mouse event state:
    state = Str( 'normal' )

    # Optional controller used for overriding event handling:
    controller = Instance( HasTraits )

    # Should debugging information be overlaid on the theme?
    debug = Bool( False )

    #-- Public Methods ---------------------------------------------------------

    def init_control ( self ):
        """ Initializes the underlying wx.Window object.
        """
        init_wx_handlers( self.control, self )

    def in_control ( self, x, y ):
        """ Returns whether a specified (x,y) coordinate is inside the control
            or not.
        """
        wdx, wdy = self.control.GetClientSize()
        return ((0 <= x < wdx) and (0 <= y < wdy))

    def refresh ( self ):
        """ Refreshes the contents of the control.
        """
        if self.control is not None:
            self.control.Refresh()

    def capture_mouse ( self ):
        """ Grab control of the mouse and indicate that we are controlling it.
        """
        if not self.control.HasCapture():
            self.control.CaptureMouse()

    def release_mouse ( self ):
        """ Grab control of the mouse and indicate that we are controlling it.
        """
        if self.control.HasCapture():
            self.control.ReleaseMouse()

    #-- Trait Event Handlers ---------------------------------------------------

    def _theme_changed ( self ):
        """ Handles the 'theme' trait being changed.
        """
        self.refresh()

    #-- wxPython Event Handlers ------------------------------------------------

    def _erase_background ( self, event ):
        """ Do not erase the background here (do it in the 'on_paint' handler).
        """
        pass

    def _paint ( self, event ):
        """ Paint the background using the associated ImageSlice object.
        """
        dc = BufferDC( self.control )
        self._paint_bg( dc )
        self._paint_fg( dc )
        dc.copy()

    def _paint_bg ( self, dc ):
        """ Paints the background into the supplied device context using the
            associated ImageSlice object and returns the image slice used (if
            any).
        """
        from image_slice import paint_parent

        # Repaint the parent's theme (if necessary):
        paint_parent( dc, self.control )

        # Draw the background theme (if any):
        if self.theme is not None:
            slice = self.theme.image_slice
            if slice is not None:
                wdx, wdy = self.control.GetClientSize()
                slice.fill( dc, 0, 0, wdx, wdy, True )

                if self.debug:
                    dc.SetPen( wx.Pen( wx.RED ) )
                    dc.SetBrush( wx.TRANSPARENT_BRUSH )
                    theme  = self.theme
                    border = theme.border
                    dc.DrawRectangle( border.left, border.top,
                                      wdx - border.right  - border.left,
                                      wdy - border.bottom - border.top )
                    dc.DrawRectangle( border.left + 3, border.top + 3,
                                      wdx - border.right  - border.left - 6,
                                      wdy - border.bottom - border.top  - 6 )
                    content = theme.content
                    x = slice.xleft + content.left
                    y = slice.xtop  + content.top
                    dc.DrawRectangle( x - 1, y - 1,
                           wdx - slice.xright  - content.right  - x + 2,
                           wdy - slice.xbottom - content.bottom - y + 2 )

                    label = theme.label
                    if slice.xtop >= slice.xbottom:
                        y, dy = 0, slice.xtop
                    else:
                        y, dy = wdy - slice.xbottom, slice.xbottom

                    if dy >= 13:
                        x  = slice.xleft + label.left
                        y += label.top
                        dc.DrawRectangle( x - 1, y - 1,
                            wdx - slice.xright - label.right - x + 2,
                            dy - label.bottom - label.top + 2 )

    def _paint_fg ( self, dc ):
        """ Paints the foreground of the window into the supplied device
            context.
        """
        pass

    def _size ( self, event ):
        """ Handles the control being resized.
        """
        control = self.control
        if control is not None:
            control.Layout()
            control.Refresh()

    def _left_down ( self, event ):
        """ Handles a left mouse button down event.
        """
        self.control.SetFocus()
        self.capture_mouse()
        self._mouse_event( 'left_down', event )

    def _left_up ( self, event ):
        """ Handles a left mouse button up event.
        """
        self.release_mouse()
        self._mouse_event( 'left_up', event )

    def _left_dclick ( self, event ):
        """ Handles a left mouse button double click event.
        """
        self.capture_mouse()
        self._mouse_event( 'left_dclick', event )

    def _middle_down ( self, event ):
        """ Handles a middle mouse button down event.
        """
        self.capture_mouse()
        self._mouse_event( 'middle_down', event )

    def _middle_up ( self, event ):
        """ Handles a middle mouse button up event.
        """
        self.release_mouse()
        self._mouse_event( 'middle_up', event )

    def _middle_dclick ( self, event ):
        """ Handles a middle mouse button double click event.
        """
        self.capture_mouse()
        self._mouse_event( 'middle_dclick', event )

    def _right_down ( self, event ):
        """ Handles a right mouse button down event.
        """
        self.capture_mouse()
        self._mouse_event( 'right_down', event )

    def _right_up ( self, event ):
        """ Handles a right mouse button up event.
        """
        self.release_mouse()
        self._mouse_event( 'right_up', event )

    def _right_dclick ( self, event ):
        """ Handles a right mouse button double click event.
        """
        self.capture_mouse()
        self._mouse_event( 'right_dclick', event )

    def _motion ( self, event ):
        """ Handles a mouse move event.
        """
        self._mouse_event( 'motion', event )

    def _enter ( self, event ):
        """ Handles the mouse entering the window event.
        """
        self._mouse_event( 'enter', event )

    def _leave ( self, event ):
        """ Handles the mouse leaving the window event.
        """
        self._mouse_event( 'leave', event )

    def _wheel ( self, event ):
        """ Handles a mouse wheel event.
        """
        self._mouse_event( 'wheel', event )

    #-- Private Methods --------------------------------------------------------

    def _mouse_event ( self, name, event ):
        """ Routes a mouse event to the proper handler (if any).
        """
        method_name = '%s_%s' % ( self.state, name )
        method      = None

        if self.controller is not None:
            method = getattr( self.controller, method_name, None )

        if method is None:
            method = getattr( self, method_name, None )

        if method is not None:
            method( event.GetX(), event.GetY(), event )




#!/usr/bin/env python

import numpy as np

import wx

import  wx.lib.buttons  as  buttons

from omnivore.utils.wx.info_panels import InfoPanel

#----------------------------------------------------------------------


from atrcopy import DefaultSegment, SegmentData

class MockEditor(object):
    def __init__(self, segment):
        self.segment = segment

    def change_bytes(self, start, end, bytes):
        print "changing bytes %d-%d to %s" % (start, end, repr(bytes))

class MockTask(object):
    def __init__(self, editor):
        self.active_editor = editor


fields = [
    ("text", "Level Number", 0x00, 2),
    ("text", "Level Name", 0x3ec, 20),
    ("colors", "Game Colors", 0x2a, 9),
    ("int", "Points per Peanut", 0x33, 2),
    ("int", "Bonus Value", 0x35, 2),
    ("int", "Number of Bullets", 0x3d, 1),
    ("int", "Number of Peanuts to Finish", 0x3e, 1),
]

class MyFrame(wx.Frame):
    def __init__(self, parent, id, title):
        wx.Frame.__init__(self, parent, id, title, wx.DefaultPosition, wx.DefaultSize)
        data = np.arange(0x800, dtype=np.uint8)
        data[0:0x50] = np.fromstring("01\x1b1\x1b1\x1b1\xa0I\xc6\x94\xaax\x0f\x00\x00B\x80E\x1b1\xe00:\xbe\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00Ws\x8e\xab\x00\x00\x86\x86\x0fT\x86\x8a\xff\xca\x00d\x00\xd0\x07\x00,F,\x80(\x01\x10L\xfdO!\x00\x00L\x00\x00 K(`\xff\x00\x8e,", dtype=np.uint8)
        data[0x3ec:0x400] = np.fromstring("abdceabdceabcdeabcde", dtype=np.uint8)
        r = SegmentData(data)
        segment = DefaultSegment(r, 0x2800)
        editor = MockEditor(segment)
        task = MockTask(editor)
        self.info = InfoPanel(self, task, fields)
        self.info.recalc_view()

class MyApp(wx.App):
    def OnInit(self):
        frame = MyFrame(None, -1, 'test_column_autosize.py')
        frame.Show(True)
        frame.Centre()
        return True

if __name__ == '__main__':
    app = MyApp(0)
    app.MainLoop()



# Enthought library imports.
from envisage.ui.tasks.api import PreferencesPane, TaskFactory
from apptools.preferences.api import PreferencesHelper
from traits.api import Bool, Int, Dict, Enum, List, Str, Unicode, Range
from traitsui.api import EnumEditor, HGroup, VGroup, Item, Label, \
    View, RangeEditor


class JumpmanPreferences(PreferencesHelper):
    """ The preferences helper for the Framework application.
    """

    #### 'PreferencesHelper' interface ########################################

    # The path to the preference node that contains the preferences.
    preferences_path = 'omnivore.tasks.jumpman_edit'

    #### Preferences ##########################################################

    # Width of bitmap in bytes
    bitmap_width_low = 1
    bitmap_width_high = 256
    bitmap_width = Range(low=bitmap_width_low, high=bitmap_width_high, value=1)


class JumpmanPreferencesPane(PreferencesPane):
    """ The preferences pane for the Framework application.
    """

    #### 'PreferencesPane' interface ##########################################

    # The factory to use for creating the preferences model object.
    model_factory = JumpmanPreferences

    category = Str('Editors')

    #### 'FrameworkPreferencesPane' interface ################################

    # See note in tasks/hex_edit/preferences.py regarding RangeEditor
    view = View(
        VGroup(HGroup(Item('bitmap_width', editor=RangeEditor(mode="spinner", is_float=False, low_name='bitmap_width_low', high_name='bitmap_width_high')),
                      Label('Default Bitmap Width (in bytes)'),
                      show_labels = False),
               label='Bitmap Editor'),
        resizable=True)



import wx

from omnivore.framework.panes import FrameworkPane

# Enthought library imports.
from pyface.api import YES, NO

# Local imports.
from disassembly import DisassemblyPanel
from segments import SegmentList
from omnivore.utils.wx.bitviewscroller import BitmapScroller, FontMapScroller, MemoryMapScroller
from omnivore.utils.wx.springtabs import SpringTabs
from omnivore.framework.undo_panel import UndoHistoryPanel
from commands import ChangeByteCommand

import logging
log = logging.getLogger(__name__)


class MemoryMapSpringTab(MemoryMapScroller):
    def activateSpringTab(self):
        self.recalc_view()

class CommentsPanel(wx.VListBox):
    def __init__(self, parent, task, **kwargs):
        self.task = task
        self.editor = None
        self.items = []
        wx.VListBox.__init__(self, parent, wx.ID_ANY, **kwargs)
        self.Bind(wx.EVT_LEFT_DOWN, self.on_click)
        
        f = self.GetFont()
        self.bold_font = wx.Font(f.GetPointSize(), f.GetFamily(),
                                 f.GetStyle(), wx.BOLD, f.GetUnderlined(),
                                 f.GetFaceName(), f.GetEncoding())
        self.italic_font = wx.Font(f.GetPointSize(), f.GetFamily(),
                                   wx.FONTSTYLE_ITALIC, wx.NORMAL, f.GetUnderlined(),
                                   f.GetFaceName(), f.GetEncoding())
        
        # Return key not sent through to EVT_CHAR, EVT_CHAR_HOOK or
        # EVT_KEY_DOWN events in a ListBox. This is the only event handler
        # that catches a Return.
        self.Bind(wx.EVT_KEY_UP, self.on_char)

    # This method must be overridden.  When called it should draw the
    # n'th item on the dc within the rect.  How it is drawn, and what
    # is drawn is entirely up to you.
    def OnDrawItem(self, dc, rect, n):
        if self.GetSelection() == n:
            c = wx.SystemSettings.GetColour(wx.SYS_COLOUR_HIGHLIGHTTEXT)
        else:
            c = self.GetForegroundColour()
        item = self.items[n]
        dc.SetFont(item[3])
        dc.SetTextForeground(c)
        dc.DrawLabel(item[2], rect,
                     wx.ALIGN_LEFT | wx.ALIGN_CENTER_VERTICAL)

    # This method must be overridden.  It should return the height
    # required to draw the n'th item.
    def OnMeasureItem(self, n):
        height = 0
        text = self.items[n][2]
        for line in text.split('\n'):
            w, h = self.GetTextExtent(line)
            height += h
        return height + 2

    def DoGetBestSize(self):
        """ Base class virtual method for sizer use to get the best size
        """
        width = 500
        height = -1
        best = wx.Size(width, height)

        # Cache the best size so it doesn't need to be calculated again,
        # at least until some properties of the window change
        self.CacheBestSize(best)

        return best
    
    def on_char(self, evt):
        keycode = evt.GetKeyCode()
        log.debug("key down: %s" % keycode)
        if keycode == wx.WXK_RETURN:
            index = self.GetSelection()
            self.process_index(index)
        evt.Skip()

    def on_click(self, evt):
        index = self.HitTest(evt.GetPosition())
        if index >= 0:
            self.SetSelection(index)
            self.process_index(index)
            self.set_items()
            self.Refresh()
    
    def process_index(self, index):
        e = self.editor
        segment, index, text, font = self.items[index]
        if segment == self.editor.segment:
            e.index_clicked(index, 0, None)
        else:
            n = e.document.find_segment_index(segment)
            e.view_segment_number(n)
            e.index_clicked(index, 0, None)
    
    def set_items(self, items=None):
        if items is None:
            items = self.input_items
        self.input_items = items
        seen = set()
        self.items = []
        for i in items:
            self.process_comment(i, seen)
        self.SetItemCount(len(self.items))
    
    def process_comment(self, item, segment_seen):
        e = self.editor
        try:
            segment = e.segment
            if segment == e.document.segments[0]:
                # if we're currently in the main segment, check for the index
                # in user segments first, rather than listing everything in
                # the main segment.
                raise IndexError
            index = segment.get_index_from_base_index(item[0])
            label = e.get_label_at_index(index)
            font = self.GetFont()
            segment_font = self.bold_font
        except IndexError:
            font = self.italic_font
            segment_font = self.italic_font
            segment, index = e.find_in_user_segment(item[0])
            if segment is not None:
                label = segment.label(index)
            else:
                index = item[0]
                label = "%04x" % index
        if segment not in segment_seen:
            segment_seen.add(segment)
            self.items.append((segment, 0, str(segment), segment_font))
        self.items.append((segment, index, "  %s: %s" % (label, item[1]), font))

    def recalc_view(self):
        e = self.task.active_editor
        self.editor = e
        if e is not None:
            self.set_items(e.document.segments[0].get_sorted_comments())
    
    def refresh_view(self):
        editor = self.task.active_editor
        if editor is not None:
            if self.editor != editor:
                self.recalc_view()
            else:
                self.Refresh()
        
    def activateSpringTab(self):
        self.recalc_view()
    
    def get_notification_count(self):
        self.recalc_view()
        return len(self.items)

class SidebarPane(FrameworkPane):
    #### TaskPane interface ###################################################

    id = 'hex_edit.sidebar'
    name = 'Sidebar'
    
    movable = False
    caption_visible = False
    dock_layer = 9
    
    def MemoryMapCB(self, parent, task, **kwargs):
        control = MemoryMapSpringTab(parent, task)
    
    def comments_cb(self, parent, task, **kwargs):
        control = CommentsPanel(parent, task)
        
    def create_contents(self, parent):
        control = SpringTabs(parent, self.task, popup_direction="left")
        control.addTab("Page Map", self.MemoryMapCB)
        control.addTab("Comments", self.comments_cb)
        return control
    
    def refresh_active(self):
        active = self.control._radio
        if active is not None and active.is_shown:
            active.managed_window.refresh_view()


class DisassemblyPane(FrameworkPane):
    #### TaskPane interface ###################################################

    id = 'hex_edit.disassembly'
    name = 'Disassembly'
    
    def create_contents(self, parent):
        control = DisassemblyPanel(parent, self.task, size=(300,500))
        return control


class BitmapPane(FrameworkPane):
    #### TaskPane interface ###################################################

    id = 'hex_edit.bitmap'
    name = 'Bitmap'
    
    def create_contents(self, parent):
        control = BitmapScroller(parent, self.task, size=(64,500))
        return control


class FontMapPane(FrameworkPane):
    #### TaskPane interface ###################################################

    id = 'hex_edit.font_map'
    name = 'Char Map'
    
    def create_contents(self, parent):
        control = FontMapScroller(parent, self.task, size=(160,500), command=ChangeByteCommand)
        return control


class SegmentsPane(FrameworkPane):
    #### TaskPane interface ###################################################

    id = 'hex_edit.segments'
    name = 'Segments'
    
    def create_contents(self, parent):
        control = SegmentList(parent, self.task, size=(64,150))
        return control


class UndoPane(FrameworkPane):
    #### TaskPane interface ###################################################

    id = 'hex_edit.undo'
    name = 'Undo History'
    
    def create_contents(self, parent):
        control = UndoHistoryPanel(parent, self.task, size=(64,150))
        return control
